'use strict';
const path = require('path');
const fs = require('fs');
const assert = require('assert');
const webpack = require('webpack');
const merge = require('webpack-merge');
const ManifestPlugin = require('webpack-manifest-plugin');
const ProgressBarPlugin = require('progress-bar-webpack-plugin');
const ImageminPlugin = require('imagemin-webpack-plugin').default;
const StatsPlugin = require('stats-webpack-plugin');
const chalk = require('chalk');
const Logger = require('./logger');
const Utils = require('../utils/utils');
const Loader = require('../utils/loader');
const defaultConfig = require('./config');

class WebpackBaseBuilder {
  constructor(config = {}) {
    this.webpack = webpack;
    this.merge = merge;
    this.options = {};
    this.loaders = [];
    this.plugins = [];
    this.packs = {};
    this.logger = new Logger(config.logger, this);
    this.initEnv(config);
    this.initConfig(config);
    this.initialize();
    this.setExtensions(['.js']);
    this.logger.env();
  }

  initOption() {
    this.createImageName();
    this.createCssName();
    this.createOption();
  }

  initLoader() {
    this.addBabelLoader();
    this.addJSONLoader();
    this.addImageLoader();
  }

  initPlugin() {
    this.addDefine({ 'process.env.NODE_ENV': JSON.stringify(process.env.NODE_ENV || 'production') });
    this.addNoError();
    this.addProgress();
  }

  prepare() {
    if (this.config.prepare) {
      this.config.prepare.bind(this);
    }
  }

  ready() {
    if (this.config.ready) {
      this.config.ready.bind(this);
    }
  }

  create() {
    this.validate();
    this.initOption();
    this.initLoader();
    this.initPlugin();
    this.ready();
    this.logger.config();
    this.logger.keyword();
    this.logger.options();
    const webpackConfig = merge(this.options, {
      module: {
        rules: this.createWebpackLoader()
      },
      plugins: this.createWebpackPlugin()
    });
    this.logger.all(webpackConfig);
    return webpackConfig;
  }

  initEnv(config) {
    // local/dev, test, prod
    this.env = config.env;
    if (this.env === 'test') {
      this.dev = false;
      this.test = true;
      this.prod = false;
    } else if (this.env === 'dev' || this.env === 'local') {
      this.dev = true;
      this.test = false;
      this.prod = false;
    } else {
      this.dev = false;
      this.test = false;
      this.prod = true;
    }
  }

  initConfig(config) {
    this.config = defaultConfig(this.env, merge(this.config, config));
  }

  initialize() {
    const config = this.config;
    this.setAlias(config.alias);
    this.setBuildPath(config.buildPath);
    this.setPublicPath(config.publicPath);
    this.setHashLength(config.hashLength);
    this.setImageHash(config.hash);
    this.setCssHash(config.hash);
    this.setMiniCss(config.miniCss);
    this.setMiniImage(config.miniImage);
    this.setMiniJs(config.miniJs);
    this.setCDN(config.cdn.url, config.cdn.dynamicDir);
    if (this.dev) {
      this.setDevMode();
    }
  }

  setHot(hot) {
    this.config.hot = hot;
  }

  setEnv(env, config) {
    this.initEnv({ env });
    this.initConfig(config);
    this.initialize();
  }

  setConfig(config) {
    this.config = merge(this.config, config);
  }

  setEntry(entry) {
    this.config.entry = entry;
  }

  setPrefix(prefix) {
    this.config.prefix = prefix;
  }

  setExtensions(extendsion) {
    this.options = merge(this.options, {
      resolve: {
        extensions: Array.isArray(extendsion) ? extendsion : [extendsion]
      }
    });
  }

  setExternals(externals) {
    this.options = merge(this.options, { externals });
  }

  setAlias(name = {}, value, useBaseDir = true) {
    if (Utils.isObject(name)) {
      Object.keys(name).forEach(key => {
        this.setAlias(key, name[key]);
      });
    } else {
      const alias = {};
      alias[name] = useBaseDir ? Utils.normalizePath(value, this.config.baseDir) : value;
      this.options = merge(this.options, {
        resolve: { alias }
      });
    }
  }

  setMiniCss(miniCss) {
    this.addPlugin(webpack.LoaderOptionsPlugin, () => Utils.isObject(miniCss) ? miniCss : { minimize: !!miniCss }, !!miniCss);
  }

  setMiniImage(miniImage) {
    this.addPlugin(ImageminPlugin, () => {
      const defaultOption = {
        pngquant: {
          quality: '90'
        }
      };
      return Utils.isObject(miniImage) ? merge(defaultOption, miniImage) : defaultOption;
    }, !!miniImage);
  }

  setMiniJs(miniJs) {
    this.addPlugin(webpack.optimize.UglifyJsPlugin, () => {
      const defaultOption = {
        compress: {
          warnings: false,
          dead_code: true,
          drop_console: true,
          drop_debugger: true
        }
      };
      return Utils.isObject(miniJs) ? merge(defaultOption, miniJs) : defaultOption;
    }, !!miniJs);
  }

  setFileName(name) {
    this.config.filename = Utils.assetsPath(this.config.prefix, name);
  }

  setHashLength(len = 8) {
    this.config.hashLength = len;
  }

  setFileNameHash(isHash) {
    this.config.fileHash = isHash;
  }

  setImageHash(isHash) {
    this.config.imageHash = isHash;
  }

  setCssHash(isHash) {
    this.config.cssHash = isHash;
  }

  setStyleLoaderName(name) {
    this.config.styleLoaderName = name;
  }

  setStyleLoaderOption(option) {
    this.config.styleLoaderOption = merge(this.config.styleLoaderOption, option);
  }

  setCommonsChunk(commonsChunk) {
    this.config.commonsChunk = Array.isArray(commonsChunk) ? commonsChunk : (commonsChunk ? [commonsChunk] : null);
  }

  addCommonsChunk(ex) {
    if (this.config.commonsChunk) {
      this.addPlugin(this.webpack.optimize.CommonsChunkPlugin, () => {
        const packKeys = Object.keys(this.packs);
        const chunks = Object.keys(this.options.entry).filter(entry => {
          return !packKeys.includes(entry);
        });
        const defaultOption = { names: this.config.commonsChunk, chunks };
        return Utils.isObject(ex) ? merge(defaultOption, ex) : defaultOption;
      }, ex);
    }
  }

  addBuildConfig() {
    if (this.config.buildConfig) {
      const filepath = path.join(this.config.baseDir, 'config/buildConfig.json');
      const buildConfig = {
        buildPath: this.buildPath.replace(this.config.baseDir, '').replace(/^\//, ''),
        publicPath: this.publicPath,
        commonsChunk: this.config.commonsChunk
      };
      Utils.saveBuildConfig(filepath, buildConfig);
    }
  }

  addManifest() {
    this.addPlugin(ManifestPlugin, () => {
      const manifest = this.config.manifest && this.config.manifest.name || 'config/manifest.json';
      const fileName = path.relative(this.buildPath, manifest);
      return { fileName };
    }, this.config.manifest);
  }

  addHot(ex) {
    this.addPlugin(this.webpack.HotModuleReplacementPlugin, null, ex);
  }

  addEntry(name, value) {
    const entry = {};
    entry[name] = value;
    this.setOption({ entry });
  }

  addBabelLoader(ex) {
    this.addLoader(/\.js$/, 'babel-loader', { exclude: /node_modules/ }, ex);
  }

  addJSONLoader(ex) {
    this.addLoader(/\.json$/, 'json-loader', ex);
  }

  addImageLoader(ex) {
    this.addLoader(/\.(png|jpe?g|gif|svg)(\?.*)?$/, 'url-loader', () => ({
      query: {
        limit: 1024,
        name: this.config.imageName
      }
    }), ex);
  }

  addDefine(defines) {
    this.addPlugin(webpack.DefinePlugin, defines, true, false);
  }

  addProvide(defines) {
    this.addPlugin(webpack.ProvidePlugin, defines, true, false);
  }

  addNoError(ex) {
    this.addPlugin(webpack.NoEmitOnErrorsPlugin, null, ex);
  }

  addProgress(ex) {
    this.addPlugin(ProgressBarPlugin, () => {
      const defaultOption = {
        width: 100,
        format: `webpack build [:bar] ${chalk.green.bold(':percent')} (:elapsed seconds)`,
        clear: false
      };
      return Utils.isObject(ex) ? merge(defaultOption, ex) : defaultOption;
    }, ex);
  }


  createEntry() {
    const configEntry = this.config.entry;
    if (configEntry && configEntry.include) {
      const entryDirs = Array.isArray(configEntry.include) ? configEntry.include : [configEntry.include];
      const normalizeEntryDirs = entryDirs.map(entryDir => Utils.normalizePath(entryDir, this.config.baseDir));
      const entry = Utils.getEntry(normalizeEntryDirs, configEntry.exclude);
      this.setOption({ entry });
    }
  }

  addStat(ex) {
    this.addPlugin(StatsPlugin, () => {
      const params = ['stats.json'];
      const defaultOption = {
        chunkModules: true,
        chunks: true,
        assets: true,
        modules: true,
        children: true,
        chunksSort: true,
        assetsSort: true
      };
      return params.push(Utils.isObject(ex) ? merge(defaultOption, ex) : defaultOption);
    }, ex);
  }

  addLoader(test, loader, option, ex) {
    option = Utils.isFunction(option) ? { fn: option } : option;
    if (Utils.isObject(test) && test.hasOwnProperty('test')) {
      loader = merge(test, option, ex);
    } else {
      loader = Utils.isString(loader) && /-loader$/.test(loader) ? require.resolve(loader) : loader;
      loader = merge({ test, loader }, option, ex);
    }
    const loaderIndex = this.findLoaderIndex(loader);
    if (loaderIndex > -1) {
      this.loaders[loaderIndex] = loader;
    } else {
      this.loaders = this.loaders.concat(loader);
    }
  }

  findLoaderIndex(loader, type) {
    return this.loaders.findIndex(item => {
      if (Utils.isString(loader)) {
        return require.resolve(loader) === require.resolve(item.loader);
      } else if (type === 'all' && loader.test && loader.loader) {
        return item.test && loader.test.toString() === item.test.toString() && loader.loader === item.loader;
      } else if (type === 'test' && loader.test) {
        return item.test && loader.test.toString() === item.test.toString();
      } else if (loader.loader) {
        return require.resolve(loader.loader) === require.resolve(item.loader);
      }
      return false;
    });
  }

  updateLoader(loader) {
    const loaderIndex = this.findLoaderIndex(loader);
    if (loaderIndex > -1) {
      this.loaders[loaderIndex] = merge(this.loaders[loaderIndex], loader);
    }
  }

  deleteLoader(loader) {
    const loaderIndex = this.findLoaderIndex(loader);
    if (loaderIndex > -1) {
      return this.loaders.splice(loaderIndex, 1);
    }
    return null;
  }

  addPlugin(clazz, args, enable, replace = true) {
    const plugin = { clazz, args, enable };
    const pluginIndex = this.findPluginIndex(clazz);
    if (replace && pluginIndex > -1) {
      this.plugins[pluginIndex] = plugin;
    } else {
      this.plugins.push(plugin);
    }
  }

  findPluginIndex(plugin) {
    const pluginName = Utils.isObject(plugin) ? plugin.constructor && plugin.constructor.name : plugin.name;
    return this.plugins.findIndex(item => {
      const configPlugin = item.clazz || item;
      const itemPluginName = Utils.isObject(configPlugin) ? configPlugin.constructor && configPlugin.constructor.name : configPlugin.name;
      return itemPluginName === pluginName;
    });
  }

  updatePlugin(plugin, args) {
    const pluginIndex = this.findPluginIndex(plugin);
    if (pluginIndex > -1) {
      this.plugins[pluginIndex] = merge(this.plugins[pluginIndex], { clazz: plugin, args });
    }
  }

  deletePlugin(plugin) {
    const pluginIndex = this.findPluginIndex(plugin);
    if (pluginIndex > -1) {
      return this.plugins.splice(pluginIndex, 1);
    }
    return null;
  }

  addPack(name = {}, value, useCommonsChunk = false) {
    if (Utils.isObject(name)) {
      Object.keys(name).forEach(packName => {
        this.addPack(packName, name[packName])
      });
    } else {
      const files = Array.isArray(value) ? value : [value];
      const normalizeFiles = files.map(file => {
        return Utils.normalizePath(file, this.config.baseDir);
      });
      this.addEntry(name, normalizeFiles);
      if (!useCommonsChunk) {
        this.packs[name] = normalizeFiles;
      }
    }
  }

  setOption(option) {
    this.options = merge(this.options, option);
  }

  setBuildPath(buildPath) {
    this.buildPath = path.isAbsolute(buildPath) ? buildPath : path.join(this.config.baseDir, buildPath);
  }

  setPublicPath(publicPath, isOverride = true) {
    if (!this.publicPath) {
      this.publicPath = '';
    }
    if (isOverride) {
      this.publicPath = publicPath;
    } else {
      this.publicPath = Utils.joinPath(this.publicPath, publicPath);
    }
    this.publicPath = `${this.publicPath.replace(/\/$/, '')}/`;
  }

  setCDN(cdnUrl, cdnDynamicDir) {
    if (cdnUrl) {
      const publicPath = cdnDynamicDir ? Utils.joinPath(cdnUrl, cdnDynamicDir) : cdnUrl;
      this.setPublicPath(publicPath);
    }
  }

  setDevMode(port) {
    this.port = port || 9000;
    const host = Utils.getHost(this.port);
    this.publicPath = Utils.joinPath(host, this.publicPath);
    const testFile = path.join(this.config.baseDir, '.webpack.js');
    if (fs.existsSync(testFile)) {
      const test = require(testFile);
      if (test.enable) {
        this.config = merge(this.config, test.config);
      }
    }
  }

  setDevTool(devtool, force) {
    if (!this.config.prod || force) {
      this.options = merge(this.options, { devtool });
    }
  }

  createWebpackLoader() {
    const webpackLoaders = [];
    const styleConfig = this.getStyleConfig();
    Loader.styleLoaders(styleConfig).forEach(loader => {
      this.loaders.push(loader);
    });
    this.loaders.forEach(loader => {
      if (loader.fn && Utils.isFunction(loader.fn)) {
        const tempLoader = Object.assign({}, loader);
        const loaderConfig = tempLoader.fn();
        delete tempLoader.fn;
        webpackLoaders.push(merge(tempLoader, loaderConfig));
      } else {
        webpackLoaders.push(loader);
      }
    });
    this.logger.loader(webpackLoaders);
    return webpackLoaders;
  }

  createWebpackPlugin() {
    const webpackPlugins = [];
    this.plugins.forEach(plugin => {
      if (plugin.enable === undefined || plugin.enable === true || (Utils.isFunction(plugin.enable) && plugin.enable())) {
        if (Utils.isObject(plugin.clazz)) {
          webpackPlugins.push(plugin.clazz);
        } else if (plugin.args) {
          const args = Utils.isFunction(plugin.args) ? plugin.args() : plugin.args;
          webpackPlugins.push(new (Function.prototype.bind.apply(plugin.clazz, [null].concat(args)))());
        } else if (!plugin.args) {
          const Clazz = plugin.clazz;
          webpackPlugins.push(new Clazz());
        }
      }
    });
    this.logger.plugin(webpackPlugins);
    return webpackPlugins;
  }

  validate() {
    assert(this.buildPath, 'webpack output path not set, please call setBuildPath method set');
    assert(this.publicPath, 'webpack output publicPath not set, please call setPublicPath method set');
    assert(this.config.entry || this.config.html || this.options.entry, 'webpack entry is empty, please call setEntry or setHtml');
  }

  createOption() {
    this.setOption({
      output: {
        path: this.buildPath,
        publicPath: this.publicPath,
        filename: this.config.filename,
        chunkFilename: this.config.chunkFilename
      }
    });
  }


  createFileName() {
    if (this.config.fileHash) {
      this.config.filename = Utils.assetsPath(this.config.prefix, `js/[name].[hash:${this.config.hashLength}].js`);
      this.config.chunkFilename = Utils.assetsPath(this.config.prefix, `js/chunk/[name].[chunkhash:${this.config.hashLength}].js`);
    } else {
      this.config.filename = Utils.assetsPath(this.config.prefix, 'js/[name].js');
      this.config.chunkFilename = Utils.assetsPath(this.config.prefix, 'js/chunk/[name].js');
    }
  }

  createImageName() {
    if (this.config.imageHash) {
      this.config.imageName = Utils.assetsPath(this.config.prefix, `img/[name].[hash:${this.config.hashLength}].[ext]`);
    } else {
      this.config.imageName = Utils.assetsPath(this.config.prefix, 'img/[name].[ext]');
    }
  }

  createCssName() {
    if (this.config.cssHash) {
      this.config.cssName = Utils.assetsPath(this.config.prefix, `css/[name].[contenthash:${this.config.hashLength}].css`);
    } else {
      this.config.cssName = Utils.assetsPath(this.config.prefix, 'css/[name].css');
    }
  }

  getStyleLoaderOption() {
    const styleConfig = this.getStyleConfig();
    return Loader.getStyleLoaderOption(styleConfig);
  }

  getStyleConfig() {
    return {
      extractCss: this.config.extractCss,
      styleLoaderName: this.config.styleLoaderName,
      styleLoaderOption: this.config.styleLoaderOption
    };
  }
}

module.exports = WebpackBaseBuilder;
